@@ -44,11 +44,6 @@ import (
 	"golang.org/x/tools/internal/versions"
 )
 
-// Various optimizations that should not affect correctness.
-const (
-	preserveImportGraph = true // hold on to the import graph for open packages
-)
-
 type unit = struct{}
 
 // A typeCheckBatch holds data for a logical type-checking operation, which may
@@ -97,21 +92,6 @@ func (b *typeCheckBatch) getHandle(id PackageID) *packageHandle {
 	return b._handles[id]
 }
 
-// A futurePackage is a future result of type checking or importing a package,
-// to be cached in a map.
-//
-// The goroutine that creates the futurePackage is responsible for evaluating
-// its value, and closing the done channel.
-type futurePackage struct {
-	done chan unit
-	v    pkgOrErr
-}
-
-type pkgOrErr struct {
-	pkg *types.Package
-	err error
-}
-
 // TypeCheck parses and type-checks the specified packages,
 // and returns them in the same order as the ids.
 // The resulting packages' types may belong to different importers,
@@ -657,7 +637,10 @@ func (b *typeCheckBatch) checkPackageForImport(ctx context.Context, ph *packageH
 	go func() {
 		exportData, err := gcimporter.IExportShallow(b.fset, pkg, bug.Reportf)
 		if err != nil {
-			bug.Reportf("exporting package %v: %v", ph.mp.ID, err)
+			// Internal error; the stack will have been reported via
+			// bug.Reportf within IExportShallow, so there's not much
+			// to do here (issue #71067).
+			event.Error(ctx, "IExportShallow failed", err, label.Package.Of(string(ph.mp.ID)))
 			return
 		}
 		if err := filecache.Set(exportDataKind, ph.key, exportData); err != nil {
@@ -701,8 +684,7 @@ func importLookup(mp *metadata.Package, source metadata.Source) func(PackagePath
 
 	// search scans children the next package in pending, looking for pkgPath.
 	// Invariant: whenever search is called, pkgPath is not yet mapped.
-	var search func(pkgPath PackagePath) (PackageID, bool)
-	search = func(pkgPath PackagePath) (id PackageID, found bool) {
+	search := func(pkgPath PackagePath) (id PackageID, found bool) {
 		pkg := pending[0]
 		pending = pending[1:]
 		for depPath, depID := range pkg.DepsByPkgPath {
@@ -1706,7 +1688,7 @@ func (b *typeCheckBatch) checkPackage(ctx context.Context, fset *token.FileSet,
 
 	// Track URIs with parse errors so that we can suppress type errors for these
 	// files.
-	unparseable := map[protocol.DocumentURI]bool{}
+	unparsable := map[protocol.DocumentURI]bool{}
 	for _, e := range pkg.parseErrors {
 		diags, err := parseErrorDiagnostics(pkg, e)
 		if err != nil {
@@ -1714,7 +1696,7 @@ func (b *typeCheckBatch) checkPackage(ctx context.Context, fset *token.FileSet,
 			continue
 		}
 		for _, diag := range diags {
-			unparseable[diag.URI] = true
+			unparsable[diag.URI] = true
 			pkg.diagnostics = append(pkg.diagnostics, diag)
 		}
 	}
@@ -1724,7 +1706,7 @@ func (b *typeCheckBatch) checkPackage(ctx context.Context, fset *token.FileSet,
 		// If the file didn't parse cleanly, it is highly likely that type
 		// checking errors will be confusing or redundant. But otherwise, type
 		// checking usually provides a good enough signal to include.
-		if !unparseable[diag.URI] {
+		if !unparsable[diag.URI] {
 			pkg.diagnostics = append(pkg.diagnostics, diag)
 		}
 	}
@@ -2001,7 +1983,7 @@ func typeErrorsToDiagnostics(pkg *syntaxPackage, inputs *typeCheckInputs, errs [
 	batch := func(related []types.Error) {
 		var diags []*Diagnostic
 		for i, e := range related {
-			code, start, end, ok := typesinternal.ReadGo116ErrorData(e)
+			code, start, end, ok := typesinternal.ErrorCodeStartEnd(e)
 			if !ok || !start.IsValid() || !end.IsValid() {
 				start, end = e.Pos, e.Pos
 				code = 0
@@ -2026,15 +2008,14 @@ func typeErrorsToDiagnostics(pkg *syntaxPackage, inputs *typeCheckInputs, errs [
 			posn := safetoken.StartPosition(e.Fset, start)
 			if !posn.IsValid() {
 				// All valid positions produced by the type checker should described by
-				// its fileset.
+				// its fileset, yet since type checker errors are associated with
+				// positions in the AST, and AST nodes can overflow the file
+				// (golang/go#48300), we can't rely on this.
 				//
-				// Note: in golang/go#64488, we observed an error that was positioned
-				// over fixed syntax, which overflowed its file. So it's definitely
-				// possible that we get here (it's hard to reason about fixing up the
-				// AST). Nevertheless, it's a bug.
-				if pkg.hasFixedFiles() {
-					bug.Reportf("internal error: type checker error %q outside its Fset (fixed files)", e)
-				} else {
+				// We should fix the parser, but in the meantime type errors are not
+				// significant if there are parse errors, so we can safely ignore this
+				// case.
+				if len(pkg.parseErrors) == 0 {
 					bug.Reportf("internal error: type checker error %q outside its Fset", e)
 				}
 				continue
@@ -2075,6 +2056,9 @@ func typeErrorsToDiagnostics(pkg *syntaxPackage, inputs *typeCheckInputs, errs [
 
 			if end == start {
 				// Expand the end position to a more meaningful span.
+				//
+				// TODO(adonovan): It is the type checker's responsibility
+				// to ensure that (start, end) are meaningful; see #71803.
 				end = analysisinternal.TypeErrorEndPos(e.Fset, pgf.Src, start)
 
 				// debugging golang/go#65960
@@ -2086,12 +2070,14 @@ func typeErrorsToDiagnostics(pkg *syntaxPackage, inputs *typeCheckInputs, errs [
 					}
 				}
 			} else {
+				// TODO(adonovan): check File(start)==File(end).
+
 				// debugging golang/go#65960
 				if _, err := safetoken.Offset(pgf.Tok, end); err != nil {
 					if pkg.hasFixedFiles() {
-						bug.Reportf("ReadGo116ErrorData returned invalid end: %v (fixed files)", err)
+						bug.Reportf("ErrorCodeStartEnd returned invalid end: %v (fixed files)", err)
 					} else {
-						bug.Reportf("ReadGo116ErrorData returned invalid end: %v", err)
+						bug.Reportf("ErrorCodeStartEnd returned invalid end: %v", err)
 					}
 				}
 			}
@@ -2143,11 +2129,11 @@ func typeErrorsToDiagnostics(pkg *syntaxPackage, inputs *typeCheckInputs, errs [
 			if i > 0 && len(diags) > 0 {
 				primary := diags[0]
 				primary.Related = append(primary.Related, protocol.DiagnosticRelatedInformation{
-					Location: protocol.Location{URI: diag.URI, Range: diag.Range},
+					Location: diag.URI.Location(diag.Range),
 					Message:  related[i].Msg, // use the unmodified secondary error for related errors.
 				})
 				diag.Related = []protocol.DiagnosticRelatedInformation{{
-					Location: protocol.Location{URI: primary.URI, Range: primary.Range},
+					Location: primary.URI.Location(primary.Range),
 				}}
 			}
 			diags = append(diags, diag)
