@@ -38,6 +38,7 @@ import (
 	gastutil "golang.org/x/tools/gopls/internal/util/astutil"
 	"golang.org/x/tools/gopls/internal/util/bug"
 	"golang.org/x/tools/gopls/internal/util/safetoken"
+	internalastutil "golang.org/x/tools/internal/astutil"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/stdlib"
 	"golang.org/x/tools/internal/tokeninternal"
@@ -138,6 +139,31 @@ func Hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, positi
 	}, nil
 }
 
+// findRhsTypeDecl finds an alias's rhs type and returns its declaration.
+// The rhs of an alias might be an alias as well, but we feel this is a rare case.
+// It returns an empty string if the given obj is not an alias.
+func findRhsTypeDecl(ctx context.Context, snapshot *cache.Snapshot, pkg *cache.Package, obj types.Object) (string, error) {
+	if alias, ok := obj.Type().(*types.Alias); ok {
+		// we choose Rhs instead of types.Unalias to make the connection between original alias
+		// and the corresponding aliased type clearer.
+		// types.Unalias brings confusion because it breaks the connection from A to C given
+		// the alias chain like 'type ( A = B; B =C ; )' except we show all transitive alias
+		// from start to the end. As it's rare, we don't do so.
+		t := alias.Rhs()
+		switch o := t.(type) {
+		case *types.Named:
+			obj = o.Obj()
+			declPGF1, declPos1, err := parseFull(ctx, snapshot, pkg.FileSet(), obj.Pos())
+			if err != nil {
+				return "", err
+			}
+			realTypeDecl, _, err := typeDeclContent(declPGF1, declPos1, obj)
+			return realTypeDecl, err
+		}
+	}
+	return "", nil
+}
+
 // hover computes hover information at the given position. If we do not support
 // hovering at the position, it returns _, nil, nil: an error is only returned
 // if the position is valid but we fail to compute hover information.
@@ -264,6 +290,10 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 		}
 	}
 
+	// By convention, we qualify hover information relative to the package
+	// from which the request originated.
+	qual := typesinternal.FileQualifier(pgf.File, pkg.Types())
+
 	// Handle hover over identifier.
 
 	// The general case: compute hover information for the object referenced by
@@ -282,10 +312,6 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 		hoverRange = &rng
 	}
 
-	// By convention, we qualify hover information relative to the package
-	// from which the request originated.
-	qual := typesinternal.FileQualifier(pgf.File, pkg.Types())
-
 	// Handle type switch identifiers as a special case, since they don't have an
 	// object.
 	//
@@ -321,6 +347,42 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 
 	// By default, types.ObjectString provides a reasonable signature.
 	signature := objectString(obj, qual, declPos, declPGF.Tok, spec)
+
+	// When hovering over a reference to a promoted struct field,
+	// show the implicitly selected intervening fields.
+	cur, ok := pgf.Cursor.FindByPos(pos, pos)
+	if !ok {
+		return protocol.Range{}, nil, fmt.Errorf("Invalid hover position, failed to get cursor")
+	}
+	if obj, ok := obj.(*types.Var); ok && obj.IsField() {
+		if selExpr, ok := cur.Parent().Node().(*ast.SelectorExpr); ok {
+			sel, ok := pkg.TypesInfo().Selections[selExpr]
+			if ok && len(sel.Index()) > 1 {
+				var buf bytes.Buffer
+				buf.WriteString(" // through ")
+				t := typesinternal.Unpointer(sel.Recv())
+				for i, index := range sel.Index()[:len(sel.Index())-1] {
+					if i > 0 {
+						buf.WriteString(", ")
+					}
+					field := typesinternal.Unpointer(t.Underlying()).(*types.Struct).Field(index)
+					t = field.Type()
+					// Inv: fieldType is N or *N for some NamedOrAlias type N.
+					if ptr, ok := t.(*types.Pointer); ok {
+						buf.WriteString("*")
+						t = ptr.Elem()
+					}
+					// Be defensive in case of ill-typed code:
+					if named, ok := t.(typesinternal.NamedOrAlias); ok {
+						buf.WriteString(named.Obj().Name())
+					}
+				}
+				// Update signature to include embedded struct info.
+				signature += buf.String()
+			}
+		}
+	}
+
 	singleLineSignature := signature
 
 	// Display struct tag for struct fields at the end of the signature.
@@ -337,6 +399,7 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 	}
 
 	// Compute size information for types,
+	// including allocator size class,
 	// and (size, offset) for struct fields.
 	//
 	// Also, if a struct type's field ordering is significantly
@@ -353,15 +416,10 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 	// use the default build config for all other types, even
 	// if they embed platform-variant types.
 	//
-	var sizeOffset string // optional size/offset description
-	// debugging #69362: unexpected nil Defs[ident] value (?)
-	_ = ident.Pos()          // (can't be nil due to check after referencedObject)
-	_ = pkg.TypesInfo()      // (can't be nil due to check in call to inferredSignature)
-	_ = pkg.TypesInfo().Defs // (can't be nil due to nature of cache.Package)
-	if def, ok := pkg.TypesInfo().Defs[ident]; ok {
-		_ = def.Pos() // can't be nil due to reasoning in #69362.
-	}
-	if def, ok := pkg.TypesInfo().Defs[ident]; ok && ident.Pos() == def.Pos() {
+	var sizeOffset string
+
+	// As painfully learned in golang/go#69362, Defs can contain nil entries.
+	if def, _ := pkg.TypesInfo().Defs[ident]; def != nil && ident.Pos() == def.Pos() {
 		// This is the declaring identifier.
 		// (We can't simply use ident.Pos() == obj.Pos() because
 		// referencedObject prefers the TypeName for an embedded field).
@@ -376,7 +434,7 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 
 		path := pathEnclosingObjNode(pgf.File, pos)
 
-		// Build string of form "size=... (X% wasted), offset=...".
+		// Build string of form "size=... (X% wasted), class=..., offset=...".
 		size, wasted, offset := computeSizeOffsetInfo(pkg, path, obj)
 		var buf strings.Builder
 		if size >= 0 {
@@ -384,6 +442,11 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 			if wasted >= 20 { // >=20% wasted
 				fmt.Fprintf(&buf, " (%d%% wasted)", wasted)
 			}
+
+			// Include allocator size class, if larger.
+			if class := sizeClass(size); class > size {
+				fmt.Fprintf(&buf, ", class=%s", format(class))
+			}
 		}
 		if offset >= 0 {
 			if buf.Len() > 0 {
@@ -404,46 +467,20 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 	_, isTypeName := obj.(*types.TypeName)
 	_, isTypeParam := types.Unalias(obj.Type()).(*types.TypeParam)
 	if isTypeName && !isTypeParam {
-		spec, ok := spec.(*ast.TypeSpec)
-		if !ok {
-			// We cannot find a TypeSpec for this type or alias declaration
-			// (that is not a type parameter or a built-in).
-			// This should be impossible even for ill-formed trees;
-			// we suspect that AST repair may be creating inconsistent
-			// positions. Don't report a bug in that case. (#64241)
-			errorf := fmt.Errorf
-			if !declPGF.Fixed() {
-				errorf = bug.Errorf
-			}
-			return protocol.Range{}, nil, errorf("type name %q without type spec", obj.Name())
+		var spec1 *ast.TypeSpec
+		typeDecl, spec1, err = typeDeclContent(declPGF, declPos, obj)
+		if err != nil {
+			return protocol.Range{}, nil, err
 		}
 
-		// Format the type's declaration syntax.
-		{
-			// Don't duplicate comments.
-			spec2 := *spec
-			spec2.Doc = nil
-			spec2.Comment = nil
-
-			var b strings.Builder
-			b.WriteString("type ")
-			fset := tokeninternal.FileSetFor(declPGF.Tok)
-			// TODO(adonovan): use a smarter formatter that omits
-			// inaccessible fields (non-exported ones from other packages).
-			if err := format.Node(&b, fset, &spec2); err != nil {
-				return protocol.Range{}, nil, err
-			}
-			typeDecl = b.String()
-
-			// Splice in size/offset at end of first line.
-			//   "type T struct { // size=..."
-			if sizeOffset != "" {
-				nl := strings.IndexByte(typeDecl, '\n')
-				if nl < 0 {
-					nl = len(typeDecl)
-				}
-				typeDecl = typeDecl[:nl] + " // " + sizeOffset + typeDecl[nl:]
+		// Splice in size/offset at end of first line.
+		//   "type T struct { // size=..."
+		if sizeOffset != "" {
+			nl := strings.IndexByte(typeDecl, '\n')
+			if nl < 0 {
+				nl = len(typeDecl)
 			}
+			typeDecl = typeDecl[:nl] + " // " + sizeOffset + typeDecl[nl:]
 		}
 
 		// Promoted fields
@@ -467,7 +504,7 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 					types.TypeString(f.field.Type(), qual),
 					f.path)
 			}
-			w.Flush()
+			w.Flush() // ignore error
 			b.WriteByte('\n')
 			fields = b.String()
 		}
@@ -478,7 +515,7 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 		// already been displayed when the node was formatted
 		// above. Don't list these again.
 		var skip map[string]bool
-		if iface, ok := spec.Type.(*ast.InterfaceType); ok {
+		if iface, ok := spec1.Type.(*ast.InterfaceType); ok {
 			if iface.Methods.List != nil {
 				for _, m := range iface.Methods.List {
 					if len(m.Names) == 1 {
@@ -520,6 +557,12 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 		}
 	}
 
+	// realTypeDecl is defined to store the underlying definition of an alias.
+	realTypeDecl, _ := findRhsTypeDecl(ctx, snapshot, pkg, obj) // tolerate the error
+	if realTypeDecl != "" {
+		typeDecl += fmt.Sprintf("\n\n%s", realTypeDecl)
+	}
+
 	// Compute link data (on pkg.go.dev or other documentation host).
 	//
 	// If linkPath is empty, the symbol is not linkable.
@@ -617,7 +660,7 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 		linkPath = ""
 	} else if linkMeta.Module != nil && linkMeta.Module.Version != "" {
 		mod := linkMeta.Module
-		linkPath = strings.Replace(linkPath, mod.Path, mod.Path+"@"+mod.Version, 1)
+		linkPath = strings.Replace(linkPath, mod.Path, cache.ResolvedString(mod), 1)
 	}
 
 	var footer string
@@ -640,6 +683,39 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 	}, nil
 }
 
+// typeDeclContent returns a well formatted type definition.
+func typeDeclContent(declPGF *parsego.File, declPos token.Pos, obj types.Object) (string, *ast.TypeSpec, error) {
+	_, spec, _ := findDeclInfo([]*ast.File{declPGF.File}, declPos) // may be nil^3
+	// Don't duplicate comments.
+	spec1, ok := spec.(*ast.TypeSpec)
+	if !ok {
+		// We cannot find a TypeSpec for this type or alias declaration
+		// (that is not a type parameter or a built-in).
+		// This should be impossible even for ill-formed trees;
+		// we suspect that AST repair may be creating inconsistent
+		// positions. Don't report a bug in that case. (#64241)
+		errorf := fmt.Errorf
+		if !declPGF.Fixed() {
+			errorf = bug.Errorf
+		}
+		return "", nil, errorf("type name %q without type spec", obj.Name())
+	}
+	spec2 := *spec1
+	spec2.Doc = nil
+	spec2.Comment = nil
+
+	var b strings.Builder
+	b.WriteString("type ")
+	fset := tokeninternal.FileSetFor(declPGF.Tok)
+	// TODO(adonovan): use a smarter formatter that omits
+	// inaccessible fields (non-exported ones from other packages).
+	if err := format.Node(&b, fset, &spec2); err != nil {
+		return "", nil, err
+	}
+	typeDecl := b.String()
+	return typeDecl, spec1, nil
+}
+
 // hoverBuiltin computes hover information when hovering over a builtin
 // identifier.
 func hoverBuiltin(ctx context.Context, snapshot *cache.Snapshot, obj types.Object) (*hoverResult, error) {
@@ -973,7 +1049,7 @@ func hoverReturnStatement(pgf *parsego.File, path []ast.Node, ret *ast.ReturnStm
 		if i > 0 {
 			buf.WriteString(", ")
 		}
-		cfg.Fprint(&buf, fset, field.Type)
+		cfg.Fprint(&buf, fset, field.Type) // ignore error
 	}
 	buf.WriteByte(')')
 	return rng, &hoverResult{
@@ -1472,16 +1548,10 @@ func findDeclInfo(files []*ast.File, pos token.Pos) (decl ast.Decl, spec ast.Spe
 	stack := make([]ast.Node, 0, 20)
 
 	// Allocate the closure once, outside the loop.
-	f := func(n ast.Node) bool {
+	f := func(n ast.Node, stack []ast.Node) bool {
 		if found {
 			return false
 		}
-		if n != nil {
-			stack = append(stack, n) // push
-		} else {
-			stack = stack[:len(stack)-1] // pop
-			return false
-		}
 
 		// Skip subtrees (incl. files) that don't contain the search point.
 		if !(n.Pos() <= pos && pos < n.End()) {
@@ -1566,7 +1636,7 @@ func findDeclInfo(files []*ast.File, pos token.Pos) (decl ast.Decl, spec ast.Spe
 		return true
 	}
 	for _, file := range files {
-		ast.Inspect(file, f)
+		internalastutil.PreorderStack(file, stack, f)
 		if found {
 			return decl, spec, field
 		}
@@ -1715,3 +1785,14 @@ func computeSizeOffsetInfo(pkg *cache.Package, path []ast.Node, obj types.Object
 
 	return
 }
+
+// sizeClass reports the size class for a struct of the specified size, or -1 if unknown.f
+// See GOROOT/src/runtime/msize.go for details.
+func sizeClass(size int64) int64 {
+	if size > 1<<16 {
+		return -1 // avoid allocation
+	}
+	// We assume that bytes.Clone doesn't trim,
+	// and reports the underlying size class; see TestSizeClass.
+	return int64(cap(bytes.Clone(make([]byte, size))))
+}
