@@ -18,6 +18,7 @@ import (
 	"go/types"
 	"io/fs"
 	"log"
+	"net/http/httptest"
 	"os"
 	"path"
 	"path/filepath"
@@ -34,6 +35,7 @@ import (
 	"golang.org/x/tools/gopls/internal/cache"
 	"golang.org/x/tools/gopls/internal/debug"
 	"golang.org/x/tools/gopls/internal/lsprpc"
+	internalmcp "golang.org/x/tools/gopls/internal/mcp"
 	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/test/compare"
 	"golang.org/x/tools/gopls/internal/test/integration"
@@ -45,6 +47,7 @@ import (
 	"golang.org/x/tools/internal/expect"
 	"golang.org/x/tools/internal/jsonrpc2"
 	"golang.org/x/tools/internal/jsonrpc2/servertest"
+	"golang.org/x/tools/internal/mcp"
 	"golang.org/x/tools/internal/testenv"
 	"golang.org/x/tools/txtar"
 )
@@ -55,7 +58,7 @@ func TestMain(m *testing.M) {
 	bug.PanicOnBugs = true
 	testenv.ExitIfSmallMachine()
 	// Disable GOPACKAGESDRIVER, as it can cause spurious test failures.
-	os.Setenv("GOPACKAGESDRIVER", "off")
+	os.Setenv("GOPACKAGESDRIVER", "off") // ignore error
 	integration.FilterToolchainPathAndGOROOT()
 	os.Exit(m.Run())
 }
@@ -96,6 +99,9 @@ func Test(t *testing.T) {
 		if strings.HasPrefix(builder, "darwin-") || strings.Contains(builder, "solaris") {
 			t.Skip("golang/go#64473: skipping with -short: this test is too slow on darwin and solaris builders")
 		}
+		if strings.HasSuffix(builder, "freebsd-amd64-race") {
+			t.Skip("golang/go#71731: the marker tests are too slow to run on the amd64-race builder")
+		}
 	}
 	// The marker tests must be able to run go/packages.Load.
 	testenv.NeedsGoPackages(t)
@@ -110,7 +116,6 @@ func Test(t *testing.T) {
 	cache := cache.New(nil)
 
 	for _, test := range tests {
-		test := test
 		t.Run(test.name, func(t *testing.T) {
 			t.Parallel()
 
@@ -133,6 +138,15 @@ func Test(t *testing.T) {
 				}
 				testenv.NeedsGo1Point(t, go1point)
 			}
+			if test.maxGoVersion != "" {
+				// A max Go version may be useful when (e.g.) a recent go/types
+				// fix makes it impossible to reproduce a certain older crash.
+				var go1point int
+				if _, err := fmt.Sscanf(test.maxGoVersion, "go1.%d", &go1point); err != nil {
+					t.Fatalf("parsing -max_go version: %v", err)
+				}
+				testenv.SkipAfterGo1Point(t, go1point)
+			}
 			if test.minGoCommandVersion != "" {
 				var go1point int
 				if _, err := fmt.Sscanf(test.minGoCommandVersion, "go1.%d", &go1point); err != nil {
@@ -174,16 +188,13 @@ func Test(t *testing.T) {
 
 			run := &markerTestRun{
 				test:       test,
-				env:        newEnv(t, cache, test.files, test.proxyFiles, test.writeGoSum, config),
+				env:        newEnv(t, cache, test.files, test.proxyFiles, test.writeGoSum, config, test.mcp),
 				settings:   config.Settings,
 				values:     make(map[expect.Identifier]any),
 				diags:      make(map[protocol.Location][]protocol.Diagnostic),
 				extraNotes: make(map[protocol.DocumentURI]map[string][]*expect.Note),
 			}
-
-			// TODO(rfindley): make it easier to clean up the integration test environment.
-			defer run.env.Editor.Shutdown(context.Background()) // ignore error
-			defer run.env.Sandbox.Close()                       // ignore error
+			defer run.env.Shutdown()
 
 			// Open all files so that we operate consistently with LSP clients, and
 			// (pragmatically) so that we have a Mapper available via the fake
@@ -286,7 +297,7 @@ func Test(t *testing.T) {
 				t.Errorf("formatTest: %v", err)
 			} else if *update {
 				filename := filepath.Join(dir, test.name)
-				if err := os.WriteFile(filename, formatted, 0644); err != nil {
+				if err := os.WriteFile(filename, formatted, 0o644); err != nil {
 					t.Error(err)
 				}
 			} else if !t.Failed() {
@@ -318,7 +329,7 @@ type marker struct {
 func (m marker) ctx() context.Context { return m.run.env.Ctx }
 
 // T returns the testing.TB for this mark.
-func (m marker) T() testing.TB { return m.run.env.T }
+func (m marker) T() testing.TB { return m.run.env.TB }
 
 // server returns the LSP server for the marker test run.
 func (m marker) editor() *fake.Editor { return m.run.env.Editor }
@@ -581,7 +592,7 @@ var actionMarkerFuncs = map[string]func(marker){
 	"highlightall":     actionMarkerFunc(highlightAllMarker),
 	"hover":            actionMarkerFunc(hoverMarker),
 	"hovererr":         actionMarkerFunc(hoverErrMarker),
-	"implementation":   actionMarkerFunc(implementationMarker),
+	"implementation":   actionMarkerFunc(implementationMarker, "err"),
 	"incomingcalls":    actionMarkerFunc(incomingCallsMarker),
 	"inlayhints":       actionMarkerFunc(inlayhintsMarker),
 	"outgoingcalls":    actionMarkerFunc(outgoingCallsMarker),
@@ -593,12 +604,15 @@ var actionMarkerFuncs = map[string]func(marker){
 	"selectionrange":   actionMarkerFunc(selectionRangeMarker),
 	"signature":        actionMarkerFunc(signatureMarker),
 	"snippet":          actionMarkerFunc(snippetMarker),
+	"subtypes":         actionMarkerFunc(subtypesMarker),
+	"supertypes":       actionMarkerFunc(supertypesMarker),
 	"quickfix":         actionMarkerFunc(quickfixMarker),
 	"quickfixerr":      actionMarkerFunc(quickfixErrMarker),
 	"symbol":           actionMarkerFunc(symbolMarker),
 	"token":            actionMarkerFunc(tokenMarker),
 	"typedef":          actionMarkerFunc(typedefMarker),
 	"workspacesymbol":  actionMarkerFunc(workspaceSymbolMarker),
+	"mcptool":          actionMarkerFunc(mcpToolMarker, "output"),
 }
 
 // markerTest holds all the test data extracted from a test txtar archive.
@@ -622,17 +636,18 @@ type markerTest struct {
 	flags      []string // flags extracted from the special "flags" archive file.
 
 	// Parsed flags values. See the flag definitions below for documentation.
-	minGoVersion        string // minimum Go runtime version; max should never be needed
-	minGoCommandVersion string
-	maxGoCommandVersion string
-	cgo                 bool
-	writeGoSum          []string
-	skipGOOS            []string
-	skipGOARCH          []string
-	ignoreExtraDiags    bool
-	filterBuiltins      bool
-	filterKeywords      bool
-	errorsOK            bool
+	minGoVersion, maxGoVersion               string // min/max version of Go runtime
+	minGoCommandVersion, maxGoCommandVersion string // min/max version of ambient go command
+
+	cgo              bool
+	writeGoSum       []string
+	skipGOOS         []string
+	skipGOARCH       []string
+	ignoreExtraDiags bool
+	filterBuiltins   bool
+	filterKeywords   bool
+	errorsOK         bool
+	mcp              bool
 }
 
 // flagSet returns the flagset used for parsing the special "flags" file in the
@@ -640,6 +655,7 @@ type markerTest struct {
 func (t *markerTest) flagSet() *flag.FlagSet {
 	flags := flag.NewFlagSet(t.name, flag.ContinueOnError)
 	flags.StringVar(&t.minGoVersion, "min_go", "", "if set, the minimum go1.X version required for this test")
+	flags.StringVar(&t.maxGoVersion, "max_go", "", "if set, the maximum go1.X version required for this test")
 	flags.StringVar(&t.minGoCommandVersion, "min_go_command", "", "if set, the minimum go1.X go command version required for this test")
 	flags.StringVar(&t.maxGoCommandVersion, "max_go_command", "", "if set, the maximum go1.X go command version required for this test")
 	flags.BoolVar(&t.cgo, "cgo", false, "if set, requires cgo (both the cgo tool and CGO_ENABLED=1)")
@@ -650,6 +666,7 @@ func (t *markerTest) flagSet() *flag.FlagSet {
 	flags.BoolVar(&t.filterBuiltins, "filter_builtins", true, "if set, filter builtins from completion results")
 	flags.BoolVar(&t.filterKeywords, "filter_keywords", true, "if set, filter keywords from completion results")
 	flags.BoolVar(&t.errorsOK, "errors_ok", false, "if set, Error level log messages are acceptable in this test")
+	flags.BoolVar(&t.mcp, "mcp", false, "if set, enable model context protocol client and server in this test")
 	return flags
 }
 
@@ -658,7 +675,7 @@ type stringListValue []string
 
 func (l *stringListValue) Set(s string) error {
 	if s != "" {
-		for _, d := range strings.Split(s, ",") {
+		for d := range strings.SplitSeq(s, ",") {
 			*l = append(*l, strings.TrimSpace(d))
 		}
 	}
@@ -777,7 +794,12 @@ func loadMarkerTest(name string, content []byte) (*markerTest, error) {
 		files:   make(map[string][]byte),
 		golden:  make(map[expect.Identifier]*Golden),
 	}
+	seen := make(map[string]bool)
 	for _, file := range archive.Files {
+		if seen[file.Name] {
+			return nil, fmt.Errorf("duplicate archive section %q", file.Name)
+		}
+		seen[file.Name] = true
 		switch {
 		case file.Name == "skip":
 			reason := strings.ReplaceAll(string(file.Data), "\n", " ")
@@ -942,7 +964,7 @@ func formatTest(test *markerTest) ([]byte, error) {
 //
 // TODO(rfindley): simplify and refactor the construction of testing
 // environments across integration tests, marker tests, and benchmarks.
-func newEnv(t *testing.T, cache *cache.Cache, files, proxyFiles map[string][]byte, writeGoSum []string, config fake.EditorConfig) *integration.Env {
+func newEnv(t *testing.T, cache *cache.Cache, files, proxyFiles map[string][]byte, writeGoSum []string, config fake.EditorConfig, enableMCP bool) *integration.Env {
 	sandbox, err := fake.NewSandbox(&fake.SandboxConfig{
 		RootDir:    t.TempDir(),
 		Files:      files,
@@ -961,26 +983,52 @@ func newEnv(t *testing.T, cache *cache.Cache, files, proxyFiles map[string][]byt
 	// Put a debug instance in the context to prevent logging to stderr.
 	// See associated TODO in runner.go: we should revisit this pattern.
 	ctx := context.Background()
-	ctx = debug.WithInstance(ctx, "off")
+	ctx = debug.WithInstance(ctx)
 
 	awaiter := integration.NewAwaiter(sandbox.Workdir)
-	ss := lsprpc.NewStreamServer(cache, false, nil)
+
+	var eventChan chan lsprpc.SessionEvent
+	var mcpServer *httptest.Server
+	if enableMCP {
+		eventChan = make(chan lsprpc.SessionEvent)
+		mcpServer = httptest.NewServer(internalmcp.HTTPHandler(eventChan, false))
+	}
+
+	ss := lsprpc.NewStreamServer(cache, false, eventChan, nil)
+
 	server := servertest.NewPipeServer(ss, jsonrpc2.NewRawStream)
 	editor, err := fake.NewEditor(sandbox, config).Connect(ctx, server, awaiter.Hooks())
 	if err != nil {
 		sandbox.Close() // ignore error
 		t.Fatal(err)
 	}
-	if err := awaiter.Await(ctx, integration.InitialWorkspaceLoad); err != nil {
+
+	if err := awaiter.Await(ctx, integration.OnceMet(
+		integration.InitialWorkspaceLoad,
+		integration.NoShownMessage(""),
+	)); err != nil {
 		sandbox.Close() // ignore error
 		t.Fatal(err)
 	}
+
+	var mcpSession *mcp.ClientSession
+	if enableMCP {
+		client := mcp.NewClient("test", "v1.0.0", nil)
+		mcpSession, err = client.Connect(ctx, mcp.NewSSEClientTransport(mcpServer.URL))
+		if err != nil {
+			t.Fatalf("fail to connect to mcp server: %v", err)
+		}
+	}
+
 	return &integration.Env{
-		T:       t,
-		Ctx:     ctx,
-		Editor:  editor,
-		Sandbox: sandbox,
-		Awaiter: awaiter,
+		TB:         t,
+		Ctx:        ctx,
+		Editor:     editor,
+		Sandbox:    sandbox,
+		Awaiter:    awaiter,
+		MCPSession: mcpSession,
+		MCPServer:  mcpServer,
+		EventChan:  eventChan,
 	}
 }
 
@@ -1029,17 +1077,17 @@ func (c *marker) sprintf(format string, args ...any) string {
 func (run *markerTestRun) fmtPos(pos token.Pos) string {
 	file := run.test.fset.File(pos)
 	if file == nil {
-		run.env.T.Errorf("position %d not in test fileset", pos)
+		run.env.TB.Errorf("position %d not in test fileset", pos)
 		return "<invalid location>"
 	}
 	m, err := run.env.Editor.Mapper(file.Name())
 	if err != nil {
-		run.env.T.Errorf("%s", err)
+		run.env.TB.Errorf("%s", err)
 		return "<invalid location>"
 	}
 	loc, err := m.PosLocation(file, pos, pos)
 	if err != nil {
-		run.env.T.Errorf("Mapper(%s).PosLocation failed: %v", file.Name(), err)
+		run.env.TB.Errorf("Mapper(%s).PosLocation failed: %v", file.Name(), err)
 	}
 	return run.fmtLoc(loc)
 }
@@ -1049,7 +1097,7 @@ func (run *markerTestRun) fmtPos(pos token.Pos) string {
 // archive file.
 func (run *markerTestRun) fmtLoc(loc protocol.Location) string {
 	if loc == (protocol.Location{}) {
-		run.env.T.Errorf("unable to find %s in test archive", loc)
+		run.env.TB.Errorf("unable to find %s in test archive", loc)
 		return "<invalid location>"
 	}
 	lines := bytes.Count(run.test.archive.Comment, []byte("\n"))
@@ -1088,12 +1136,12 @@ func (run *markerTestRun) mapLocation(loc protocol.Location) (name string, start
 	name = run.env.Sandbox.Workdir.URIToPath(loc.URI)
 	m, err := run.env.Editor.Mapper(name)
 	if err != nil {
-		run.env.T.Errorf("internal error: %v", err)
+		run.env.TB.Errorf("internal error: %v", err)
 		return
 	}
 	start, end, err := m.RangeOffsets(loc.Range)
 	if err != nil {
-		run.env.T.Errorf("error formatting location %s: %v", loc, err)
+		run.env.TB.Errorf("error formatting location %s: %v", loc, err)
 		return
 	}
 	startLine, startCol = m.OffsetLineCol8(start)
@@ -1366,7 +1414,6 @@ func (sm stringMatcher) check(mark marker, got string) {
 		if !sm.pattern.MatchString(got) {
 			mark.errorf("got %q, does not match pattern %#q", got, sm.pattern)
 		}
-
 	} else if !strings.Contains(got, sm.substr) {
 		// Content must contain the expected substring.
 		mark.errorf("got %q, want substring %q", got, sm.substr)
@@ -1380,7 +1427,6 @@ func checkChangedFiles(mark marker, changed map[string][]byte, golden *Golden) {
 		if want, ok := golden.Get(mark.T(), filename, got); !ok {
 			mark.errorf("%s: unexpected change to file %s; got:\n%s",
 				mark.note.Name, filename, got)
-
 		} else if string(got) != string(want) {
 			mark.errorf("%s: wrong file content for %s: got:\n%s\nwant:\n%s\ndiff:\n%s",
 				mark.note.Name, filename, got, want,
@@ -1433,7 +1479,6 @@ func checkDiffs(mark marker, changed map[string][]byte, golden *Golden) {
 		if want, ok := golden.Get(mark.T(), filename, []byte(got)); !ok {
 			mark.errorf("%s: unexpected change to file %s; got diff:\n%s",
 				mark.note.Name, filename, got)
-
 		} else if got != string(want) {
 			mark.errorf("%s: wrong diff for %s:\n\ngot:\n%s\n\nwant:\n%s\n",
 				mark.note.Name, filename, got, want)
@@ -1584,7 +1629,7 @@ func completeMarker(mark marker, src protocol.Location, want ...completionItem)
 		want = nil // got is nil if empty
 	}
 	if diff := cmp.Diff(want, got); diff != "" {
-		mark.errorf("Completion(...) returned unexpect results (-want +got):\n%s", diff)
+		mark.errorf("Completion(...) returned unexpected results (-want +got):\n%s", diff)
 	}
 }
 
@@ -1632,7 +1677,6 @@ func acceptCompletionMarker(mark marker, src protocol.Location, label string, go
 	filename := mark.path()
 	mapper := mark.mapper()
 	patched, _, err := protocol.ApplyEdits(mapper, append([]protocol.TextEdit{edit}, selected.AdditionalTextEdits...))
-
 	if err != nil {
 		mark.errorf("ApplyProtocolEdits failed: %v", err)
 		return
@@ -1643,20 +1687,22 @@ func acceptCompletionMarker(mark marker, src protocol.Location, label string, go
 }
 
 // defMarker implements the @def marker, running textDocument/definition at
-// the given src location and asserting that there is exactly one resulting
-// location, matching dst.
-//
-// TODO(rfindley): support a variadic destination set.
-func defMarker(mark marker, src, dst protocol.Location) {
-	got := mark.run.env.GoToDefinition(src)
-	if got != dst {
-		mark.errorf("definition location does not match:\n\tgot: %s\n\twant %s",
-			mark.run.fmtLoc(got), mark.run.fmtLoc(dst))
+// the given location and asserting that there the results match want.
+func defMarker(mark marker, loc protocol.Location, want ...protocol.Location) {
+	env := mark.run.env
+	got, err := env.Editor.Definitions(env.Ctx, loc)
+	if err != nil {
+		mark.errorf("definition request failed: %v", err)
+		return
+	}
+
+	if err := compareLocations(mark, got, want); err != nil {
+		mark.errorf("def failed: %v", err)
 	}
 }
 
 func typedefMarker(mark marker, src, dst protocol.Location) {
-	got := mark.run.env.TypeDefinition(src)
+	got := mark.run.env.FirstTypeDefinition(src)
 	if got != dst {
 		mark.errorf("type definition location does not match:\n\tgot: %s\n\twant %s",
 			mark.run.fmtLoc(got), mark.run.fmtLoc(dst))
@@ -1684,8 +1730,9 @@ func foldingRangeMarker(mark marker, g *Golden) {
 		})
 	}
 	for i, rng := range ranges {
-		insert(rng.StartLine, rng.StartCharacter, fmt.Sprintf("<%d kind=%q>", i, rng.Kind))
-		insert(rng.EndLine, rng.EndCharacter, fmt.Sprintf("</%d>", i))
+		// We assume the server populates these optional fields.
+		insert(*rng.StartLine, *rng.StartCharacter, fmt.Sprintf("<%d kind=%q>", i, rng.Kind))
+		insert(*rng.EndLine, *rng.EndCharacter, fmt.Sprintf("</%d>", i))
 	}
 	filename := mark.path()
 	mapper, err := env.Editor.Mapper(filename)
@@ -1748,6 +1795,7 @@ func highlightLocationMarker(mark marker, loc protocol.Location, kindName expect
 		Kind:  kind,
 	}
 }
+
 func sortDocumentHighlights(s []protocol.DocumentHighlight) {
 	sort.Slice(s, func(i, j int) bool {
 		return protocol.CompareRange(s[i].Range, s[j].Range) < 0
@@ -1766,7 +1814,7 @@ func sortDocumentHighlights(s []protocol.DocumentHighlight) {
 func highlightAllMarker(mark marker, all ...protocol.DocumentHighlight) {
 	sortDocumentHighlights(all)
 	for _, src := range all {
-		loc := protocol.Location{URI: mark.uri(), Range: src.Range}
+		loc := mark.uri().Location(src.Range)
 		got := mark.run.env.DocumentHighlight(loc)
 		sortDocumentHighlights(got)
 
@@ -1777,7 +1825,7 @@ func highlightAllMarker(mark marker, all ...protocol.DocumentHighlight) {
 }
 
 func highlightMarker(mark marker, src protocol.DocumentHighlight, dsts ...protocol.DocumentHighlight) {
-	loc := protocol.Location{URI: mark.uri(), Range: src.Range}
+	loc := mark.uri().Location(src.Range)
 	got := mark.run.env.DocumentHighlight(loc)
 
 	sortDocumentHighlights(got)
@@ -1811,7 +1859,7 @@ func locMarker(mark marker, loc protocol.Location) protocol.Location { return lo
 // defLocMarker implements the @defloc marker, which binds a location to the
 // (first) result of a jump-to-definition request.
 func defLocMarker(mark marker, loc protocol.Location) protocol.Location {
-	return mark.run.env.GoToDefinition(loc)
+	return mark.run.env.FirstDefinition(loc)
 }
 
 // diagMarker implements the @diag marker. It eliminates diagnostics from
@@ -1835,7 +1883,7 @@ func removeDiagnostic(mark marker, loc protocol.Location, matchEnd bool, re *reg
 	diags := mark.run.diags[key]
 	for i, diag := range diags {
 		if re.MatchString(diag.Message) && (!matchEnd || diag.Range.End == loc.Range.End) {
-			mark.run.diags[key] = append(diags[:i], diags[i+1:]...)
+			mark.run.diags[key] = slices.Delete(diags, i, i+1)
 			return diag, true
 		}
 	}
@@ -1937,8 +1985,12 @@ func signatureMarker(mark marker, src protocol.Location, label string, active in
 	if got := gotLabels[0]; got != label {
 		mark.errorf("signatureHelp: got label %q, want %q", got, label)
 	}
-	if got := int64(got.ActiveParameter); got != active {
-		mark.errorf("signatureHelp: got active parameter %d, want %d", got, active)
+	gotActiveParameter := int64(-1) // => missing
+	if got.ActiveParameter != nil {
+		gotActiveParameter = int64(*got.ActiveParameter)
+	}
+	if gotActiveParameter != active {
+		mark.errorf("signatureHelp: got active parameter %d, want %d", gotActiveParameter, active)
 	}
 }
 
@@ -2147,7 +2199,7 @@ func documentLinkMarker(mark marker, g *Golden) {
 			mark.errorf("%s: nil link target", l.Range)
 			continue
 		}
-		loc := protocol.Location{URI: mark.uri(), Range: l.Range}
+		loc := mark.uri().Location(l.Range)
 		fmt.Fprintln(&b, mark.run.fmtLocForGolden(loc), *l.Target)
 	}
 
@@ -2300,11 +2352,11 @@ func codeActionChanges(env *integration.Env, uri protocol.DocumentURI, rng proto
 
 	if action.Edit != nil {
 		if len(action.Edit.Changes) > 0 {
-			env.T.Errorf("internal error: discarding unexpected CodeAction{Kind=%s, Title=%q}.Edit.Changes", action.Kind, action.Title)
+			env.TB.Errorf("internal error: discarding unexpected CodeAction{Kind=%s, Title=%q}.Edit.Changes", action.Kind, action.Title)
 		}
 		if action.Edit.DocumentChanges != nil {
 			if action.Command != nil {
-				env.T.Errorf("internal error: discarding unexpected CodeAction{Kind=%s, Title=%q}.Command", action.Kind, action.Title)
+				env.TB.Errorf("internal error: discarding unexpected CodeAction{Kind=%s, Title=%q}.Command", action.Kind, action.Title)
 			}
 			return action.Edit.DocumentChanges, nil
 		}
@@ -2369,22 +2421,63 @@ func refsMarker(mark marker, src protocol.Location, want ...protocol.Location) {
 
 // implementationMarker implements the @implementation marker.
 func implementationMarker(mark marker, src protocol.Location, want ...protocol.Location) {
+	wantErr := namedArgFunc(mark, "err", convertStringMatcher, stringMatcher{})
+
 	got, err := mark.server().Implementation(mark.ctx(), &protocol.ImplementationParams{
 		TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(src),
 	})
-	if err != nil {
+	if err != nil && wantErr.empty() {
 		mark.errorf("implementation at %s failed: %v", src, err)
 		return
 	}
+	if !wantErr.empty() {
+		wantErr.checkErr(mark, err)
+		return
+	}
 	if err := compareLocations(mark, got, want); err != nil {
 		mark.errorf("implementation: %v", err)
 	}
 }
 
-func itemLocation(item protocol.CallHierarchyItem) protocol.Location {
-	return protocol.Location{
-		URI:   item.URI,
-		Range: item.Range,
+func mcpToolMarker(mark marker, tool string, rawArgs string, loc protocol.Location) {
+	args := make(map[string]any)
+	if err := json.Unmarshal([]byte(rawArgs), &args); err != nil {
+		mark.errorf("fail to unmarshal arguments to map[string]any: %v", err)
+		return
+	}
+
+	// TODO(hxjiang): Make the "location" key configurable.
+	args["location"] = loc
+
+	res, err := mark.run.env.MCPSession.CallTool(mark.ctx(), &mcp.CallToolParams{
+		Name:      tool,
+		Arguments: args,
+	})
+	if err != nil {
+		mark.errorf("failed to call mcp tool: %v", err)
+		return
+	}
+
+	var buf bytes.Buffer
+	for i, c := range res.Content {
+		if c.Type != "text" {
+			mark.errorf("unsupported return content[%v] type: %s", i, c.Type)
+			continue
+		}
+		buf.WriteString(c.Text)
+	}
+	if !bytes.HasSuffix(buf.Bytes(), []byte{'\n'}) {
+		buf.WriteString("\n") // all golden content is newline terminated
+	}
+
+	got := buf.String()
+
+	output := namedArg(mark, "output", expect.Identifier(""))
+	golden := mark.getGolden(output)
+	if want, ok := golden.Get(mark.T(), "", []byte(got)); !ok {
+		mark.errorf("%s: missing golden file @%s", mark.note.Name, golden.id)
+	} else if diff := cmp.Diff(got, string(want)); diff != "" {
+		mark.errorf("unexpected mcp tools call %s return; got:\n%s\n want:\n%s\ndiff:\n%s", tool, got, want, diff)
 	}
 }
 
@@ -2396,7 +2489,7 @@ func incomingCallsMarker(mark marker, src protocol.Location, want ...protocol.Lo
 		}
 		var locs []protocol.Location
 		for _, call := range calls {
-			locs = append(locs, itemLocation(call.From))
+			locs = append(locs, call.From.URI.Location(call.From.Range))
 		}
 		return locs, nil
 	}
@@ -2411,7 +2504,7 @@ func outgoingCallsMarker(mark marker, src protocol.Location, want ...protocol.Lo
 		}
 		var locs []protocol.Location
 		for _, call := range calls {
-			locs = append(locs, itemLocation(call.To))
+			locs = append(locs, call.To.URI.Location(call.To.Range))
 		}
 		return locs, nil
 	}
@@ -2432,7 +2525,8 @@ func callHierarchy(mark marker, src protocol.Location, getCalls callHierarchyFun
 		mark.errorf("PrepareCallHierarchy returned %d items, want exactly 1", nitems)
 		return
 	}
-	if loc := itemLocation(items[0]); loc != src {
+	item := items[0]
+	if loc := item.URI.Location(item.Range); loc != src {
 		mark.errorf("PrepareCallHierarchy found call %v, want %v", loc, src)
 		return
 	}
@@ -2442,16 +2536,8 @@ func callHierarchy(mark marker, src protocol.Location, getCalls callHierarchyFun
 		return
 	}
 	if calls == nil {
-		calls = []protocol.Location{}
+		calls = []protocol.Location{} // non-nil; cmp.Diff cares
 	}
-	// TODO(rfindley): why aren't call hierarchy results stable?
-	sortLocs := func(locs []protocol.Location) {
-		sort.Slice(locs, func(i, j int) bool {
-			return protocol.CompareLocation(locs[i], locs[j]) < 0
-		})
-	}
-	sortLocs(want)
-	sortLocs(calls)
 	if d := cmp.Diff(want, calls); d != "" {
 		mark.errorf("call hierarchy: unexpected results (-want +got):\n%s", d)
 	}
@@ -2514,6 +2600,50 @@ func prepareRenameMarker(mark marker, src protocol.Location, placeholder string)
 	}
 }
 
+func subtypesMarker(mark marker, src protocol.Location, want ...protocol.Location) {
+	typeHierarchy(mark, src, want, func(item protocol.TypeHierarchyItem) ([]protocol.TypeHierarchyItem, error) {
+		return mark.server().Subtypes(mark.ctx(), &protocol.TypeHierarchySubtypesParams{Item: item})
+	})
+}
+
+func supertypesMarker(mark marker, src protocol.Location, want ...protocol.Location) {
+	typeHierarchy(mark, src, want, func(item protocol.TypeHierarchyItem) ([]protocol.TypeHierarchyItem, error) {
+		return mark.server().Supertypes(mark.ctx(), &protocol.TypeHierarchySupertypesParams{Item: item})
+	})
+}
+
+type typeHierarchyFunc = func(item protocol.TypeHierarchyItem) ([]protocol.TypeHierarchyItem, error)
+
+func typeHierarchy(mark marker, src protocol.Location, want []protocol.Location, get typeHierarchyFunc) {
+	items, err := mark.server().PrepareTypeHierarchy(mark.ctx(), &protocol.TypeHierarchyPrepareParams{
+		TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(src),
+	})
+	if err != nil {
+		mark.errorf("PrepareTypeHierarchy failed: %v", err)
+		return
+	}
+	if nitems := len(items); nitems != 1 {
+		mark.errorf("PrepareTypeHierarchy returned %d items, want exactly 1", nitems)
+		return
+	}
+	if loc := (protocol.Location{URI: items[0].URI, Range: items[0].Range}); loc != src {
+		mark.errorf("PrepareTypeHierarchy found type %v, want %v", loc, src)
+		return
+	}
+	items, err = get(items[0])
+	if err != nil {
+		mark.errorf("type hierarchy failed: %v", err)
+		return
+	}
+	got := []protocol.Location{} // non-nil; cmp.Diff cares
+	for _, item := range items {
+		got = append(got, item.URI.Location(item.Range))
+	}
+	if d := cmp.Diff(want, got); d != "" {
+		mark.errorf("type hierarchy: unexpected results (-want +got):\n%s", d)
+	}
+}
+
 // symbolMarker implements the @symbol marker.
 func symbolMarker(mark marker, golden *Golden) {
 	// Retrieve information about all symbols in this file.
